// recoderc: a simple client using mplayer 
//           to present a MPEG stream from recoderd
//
// (c) 2004 Patrick Cernko <errror@gmx.de>

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <unistd.h>    // pipe(), close(), read(), write(), exit(), execl(), select()
#include <fcntl.h>     // fcntl()
#include <sys/time.h>  // select()
#include <sys/types.h> // select()
#include <sys/wait.h>  // waitpid()
#include <errno.h>     // errno
#include <signal.h>    // SIGTERM

#include <iostream>
#include <string>

// includes from the toolbox lib
#include "cfile.h"
#include "csocket.h"
#include "cselect.h"


#ifndef DEFAULT_CMD_PORT
#  define DEFAULT_CMD_PORT 40404
#endif

#ifndef RECODERD_HOSTNAME
#  define RECODERD_HOSTNAME "192.168.149.2"
#endif

#ifndef LOCAL_MP_SKIP_FRAMES
#  define LOCAL_MP_SKIP_FRAMES 5
#endif

using namespace std;

int main(int argc, char* argv[]) {
  int cmd_port(DEFAULT_CMD_PORT);
  string recoderd_hostname(RECODERD_HOSTNAME);
  string filename("");
  int data_port(0);

  // parse args
  if (argc < 2) {
    cerr << "Usage:" << endl
         << "  " << argv[0] << " <filename>" << endl
         << endl;
    return 1;
  }
  filename = string(argv[1]);

  // used objects
  cTBSocket cmd;             // a soccket for connecting to server
  cTBSocket data;            // a soccket for receiving the trancoded MPEG stream
  cTBSocket net_listener;    // a socket listener
  cTBSelect io_select;       // a select() object

  int inpipe[2], outpipe[2]; // pipes for the std{in,out,err} connection to mplayer
  int pid(0);                // the pid of our mplayer child

  cTBFile mp_input;          // mplayer's input is a output for us
  cTBFile mp_output;         // mplayer's output is a input for us
  cTBFile std_input;         // stdin as TB-object for TB's select
  cTBFile fifo_output;       // mplayer's data input via the fifo


  // create a fifo for the data output from mplayer
  char tmp_dir_tmpl[25];
  strncpy(tmp_dir_tmpl, "/tmp/recoderd.XXXXXX", 25);

#ifdef HAVE_MKDTEMP
  char* tmp_dir(mkdtemp(tmp_dir_tmpl));
#else
#  ifdef HAVE_MKTEMP
  char* tmp_dir(mktemp(tmp_dir_tmpl));
#  else
#    error no usable implemtation for creating tmp files/dirs available!
#  endif
#endif
  if (tmp_dir == NULL) {
    cerr << "ERROR: Could not create temp. dir: " << strerror(errno) << endl;
    return 1;
  }
#if !defined(HAVE_MKDTEMP) && defined(HAVE_MKTEMP)
  mkdir(tmp_dir, 00777);
#endif

  string fifoname(tmp_dir);
  fifoname.append("/file.mpeg");
  if (mkfifo(fifoname.c_str(), 0600) == -1) {
    cerr << "ERROR: Problems creating fifo \"" << fifoname << "\": " << strerror(errno) << endl;
    unlink("grab.mpg");
    rmdir(tmp_dir);
    return 1;
  }

  // connect to server
  if (!cmd.Connect(recoderd_hostname.c_str(), cmd_port)) {
    cerr << "ERROR: Could not connect to " << recoderd_hostname << ":" << cmd_port << endl;
    return 1;
  }



  // send request
  // 1. wait for write possibility
  do {
    io_select.Clear();
    io_select.Add(cmd, true /* waiting for possibility to output */);
    io_select.Select(100);
  } while (!io_select.CanWrite(cmd));
  // 2. write request: the filename
  char* request_data((char*) (filename + string("\n")).c_str());
  size_t request_size(filename.length() + 1);
  while (request_size > 0) {
    cmd.Write(request_data, request_size);
  }
  // :XXX: todo: making sure that all data is written



  // read answer: the data port
  string data_port_string("");
  do {
    // 1. wait for incoming data
    do {
      io_select.Clear();
      io_select.Add(cmd, false /* waiting for input*/);
      io_select.Select(100); // wait 100ms
    } while (!io_select.CanRead(cmd));

    // 2. :XXX: todo: error: connection closed?

    // 3. read the data from the socket
    char buffer[2048];
    size_t bytes_read(0);
    if (cmd.Read(buffer, 2048, bytes_read) != -1) {
      // 4. if we did not have an error ...
      // append read data to filename string
      data_port_string.append(string(buffer, bytes_read));
    }
    // 5. do this until we read a newline character
  } while (data_port_string[data_port_string.length()-1] != '\n');
  // 6. chomp() data_port_string
  while ((data_port_string[data_port_string.length()-1] == '\n') ||
         (data_port_string[data_port_string.length()-1] == '\r')) {
    data_port_string = data_port_string.substr(0, data_port_string.length()-1);
  }
  // 7. convert data_port_string to int
  data_port = atoi(data_port_string.c_str());
  if (data_port == 0) {
    cerr << "ERROR: Could not parse port for data connection: " << data_port_string << endl;
    return 1;
  }



  // connect to data port
  if (!data.Connect(recoderd_hostname.c_str(), data_port)) {
    cerr << "ERROR: Could not connect to " << recoderd_hostname << ":" << data_port << endl;
    return 1;
  }



  // now we have all connection via network established:
  // 1. we have connected to recoder_hostname:cmd_port and send the filename
  // 3. we received the data port number of the server as answer for our request
  // 4. we have connected to the data port of the server
  


  // open/create the pipes for mplayer
  // :XXX: this should be done at program start to avoid habing errors here
  if (pipe(inpipe) == -1) {
    cerr << "ERROR: pipe failed for inpipe: (" << errno << ") " << strerror(errno) << endl;
    return 1;
  }
  if (pipe(outpipe) == -1) {
    cerr << "ERROR: pipe failed for outpipe: (" << errno << ") " << strerror(errno) << endl;
    return 1;
  }


  // fork a child
  pid = fork();
  if (pid == -1) {
    cerr << "ERROR: fork failed: (" << errno << ") " << strerror(errno) << endl;
    return 1;
  }


  // the child: exec a mplayer
  if (pid == 0) {
    // 1. close other side's pipe ends
    close(inpipe[1]);
    close(outpipe[0]);

#ifdef DEBUG
    cerr << "mplayer" << " "
         << "-aspect" << " " << "4:3" << " "
         << "-nocache" << " "
         << "-demuxer" << " " << "2" << " "
         << "-slave" << " "
         << "-really-quiet" << " "
         << "-osdlevel" << " " << "0" << " "
#ifndef DEBUG_LOCAL
         << "-vo" << " " << "mpegpes" << " "
         << "-ao" << " " << "mpegpes" << " "
#endif // !DEBUG_LOCAL
         << fifoname.c_str() << endl;
#endif // DEBUG

    // 2. dup stdio to this side's pipe ends
    if((dup2(inpipe[0],  STDIN_FILENO)  < 0) ||
       (dup2(outpipe[1], STDOUT_FILENO) < 0) ||
       (dup2(outpipe[1], STDERR_FILENO) < 0)) {
      cerr << "ERROR: dup2() failed in MPlayer child: (" << errno << ") " << strerror(errno) << endl;
      exit(127);
    }
    // 3. close this side's pipes as the are dup'ed to std-io
    close(inpipe[0]);
    close(outpipe[1]);

    // 4. the final command
    execlp("mplayer",
           "mplayer",
           "-aspect", "4:3",
           "-nocache",
           "-demuxer", "2",
           "-slave",
           "-quiet",
           "-osdlevel", "0",
#ifndef DEBUG_LOCAL
           "-vo", "mpegpes",
           "-ao", "mpegpes",
#endif
           fifoname.c_str(),
           NULL);

    // 5. error handling
    cerr << "Couldn't start mplayer: " << strerror(errno) << endl;
    exit(127);
  }
  // end of child code



  // the father
  //
  // if we do not sleep some time, the forked 
  // child hangs, don't know why
  usleep(50000);



  // close other side's pipe ends
  close(inpipe[0]);
  close(outpipe[1]);



  // open/create TB-objects for the communication with mplayer
  mp_input.Open(inpipe[1]);
  mp_output.Open(outpipe[0]);
  // ... and stdin
  std_input.Open(STDIN_FILENO);


  bool local_mp_running = false;
  unsigned int local_mp_skip_frames = LOCAL_MP_SKIP_FRAMES;
  // main loop
  // 1. as long as our mplayer child is alive
  while (waitpid(pid, 0, WNOHANG) != pid) {
    // 2. build up a select
    // 2.1 clear all
    io_select.Clear();
    // 2.2 add mplayer's output (INPUT for us!)
    io_select.Add(mp_output, false /* waiting for input! */);
    // 2.3 add net data output (INPUT for us!)
    io_select.Add(data, false /* waiting for input! */);
    if (local_mp_running) {
      // 2.4 add std input
      io_select.Add(std_input, false /* waiting for input! */);
    }
    // 2.5 add net cmd output (INPUT for us!)
    io_select.Add(cmd, false);
    // 3. select the descriptors
    if (io_select.Select(200) > 0) {

      // at least one of the has some data
      // 1. most important is mplayer's output
      //    as if it is blocked, this can lead to a deadlock!
      if (io_select.CanRead(mp_output)) {
        // 1.1 read what mplayer wants us to tell!
        char mplayer_output[1025];
        size_t bytes_read(0);
        // 1.2 if we did not have an error ...
        if (mp_output.Read(mplayer_output, 1024, bytes_read) != -1) {
          // make sure, string is terminated correctly
          mplayer_output[bytes_read] = '\0';
          if (!local_mp_running && strstr(mplayer_output, "SLAVE: time") != NULL) {
            local_mp_running = (--local_mp_skip_frames == 0);
#ifdef DEBUG
            cerr << endl << endl << "Found the needle in the haystack!" << endl << endl;
#endif
          }
          // ... write what we've read
//           cout.write(mplayer_output, bytes_read);
//           cout.flush();
          // to /dev/null :-)
        }
      }

      // 2. next important is remote cmd's
      if (io_select.CanRead(cmd)) {
        // 1.1 read what mplayer wants us to tell!
        char mplayer_output[1024];
        size_t bytes_read(0);
        if (cmd.Read(mplayer_output, 1024, bytes_read) != -1) {
          // 1.2 if we did not have an error ...
          if (0 == bytes_read) {
            size_t quit_size(5);
            mp_input.Write("quit\n", quit_size);
            break;
          } else {
            // ... write what we've read
            cout.write(mplayer_output, bytes_read);
            cout.flush();
          }
        }
      }

      string next_command("");

      // 2. next important: commands from stdin
      if (local_mp_running && io_select.CanRead(std_input)) {
        // 2.1 read the command
        char buf[256];
        size_t read_bytes(0);
        std_input.Read(buf, 256, read_bytes);

        // 2.2 put it in the next-command string
        next_command = string(buf, read_bytes);
        // 2.3 find newline in the string
        if (next_command.find('\n') == string::npos) {
          // ... if not: ignore command
          // :XXX: to buffered reading over multiple tries
          next_command = string("");
        } else {
          // chomp() the string
          while ((next_command[next_command.length()-1] == '\n') ||
                 (next_command[next_command.length()-1] == '\r')) {
            next_command = next_command.substr(0, next_command.length()-1);
          }
        }
      }

      // 3. if we have got a next command
      if (!next_command.empty()) {
        next_command.append("\n");
        char* cmd_buffer((char*) next_command.c_str());
        size_t cmd_buffer_length(next_command.length());

        cTBFileNo* cmd_target(NULL);
        if ((next_command.find("volume") != string::npos) ||
            (next_command.find("mute") != string::npos)) {
          // volume must be set local!
          cmd_target = &mp_input;
        } else {
          cmd_target = &cmd;
        }

        // write it to net cmd
        while (cmd_buffer_length > 0) {
          cmd_target->Write(cmd_buffer, cmd_buffer_length);
        }
      }

      // 4. if net data has some data for us
      if (io_select.CanRead(data)) {
        // 5.1 read the data
        char buffer[4096];
        size_t bytes_read(0);
        data.Read(buffer, 4096, bytes_read);
        if (0 == bytes_read) {
          size_t quit_size(5);
          mp_input.Write("quit\n", quit_size);
          break;
        }
        // 5.2 and write it to the data network connection
        if (!fifo_output.IsOpen()) {
          fifo_output.Open(fifoname.c_str(), O_WRONLY + O_NONBLOCK);
        }
        while (bytes_read > 0) {
          // all of it!
          do {
            io_select.Clear();
            io_select.Add(fifo_output, true /* waiting for possibility to output */);
            io_select.Select(100);
          } while (!io_select.CanWrite(fifo_output));
          fifo_output.Write(buffer, bytes_read);
        }
      }

    } // end if: data available on one of the connections
  } // end while: main loop

  fifo_output.Close();

  // shutdown
  // 1. wait until mplayer child died (otherwise we get a zoobie)
  if (waitpid(pid, 0, WNOHANG) != -1) {
    // child not terminated yet
    usleep(100000); // wait 100ms
//     cerr << "If it won't die";
    // as long as it did not terminate
    while (waitpid(pid, 0, WNOHANG) == 0) {
      usleep(100000); // wait 100ms
//       cerr << ", we'll shoot it";
      // and KILL it!
      kill(pid, SIGTERM);
    }
//     cerr << endl;
  }


//   cerr << "exit" << endl;

  return 0;
}
